; The Midnight Markup Masher - an attempt at world domination in Scheme
; ICFP 2001 entry by John R. Hall <overcode@cc.gatech.edu>

; Scans the next text string out of a string of SML.
; Stops at the first tag opening encountered.
; Returns two values: the scanned string, and the remaining SML string.
(define sml-read-text
  (lambda (str)
    (let ((len (string-length str)))
      (let reader ((pos 0))
        (if (or (= pos len)
                (equal? (string-ref str pos) #\<))
            (values (substring str 0 pos)
                    (substring str pos len))
            (reader (+ pos 1)))))))

; Scans the next tag out of a string of SML.
; Stops at the end of the tag.
; Returns two values: the scanned tag, and the remaining SML string.
(define sml-read-tag
  (lambda (str)
    (let ((len (string-length str)))
      (let reader ((pos 0))
        (if (or (= pos len)
                (equal? (string-ref str pos) #\>))
            (let ((s (if (< pos len) (+ pos 1) pos)))
              (values (substring str 0 s)
                      (substring str s len)))
            (reader (+ pos 1)))))))

; Scans the next token (tag or text) out of a string of SML.
; Returns two values: the scanned token, and the remaining SML string.
(define sml-read-token
  (lambda (str)
    (if (= (string-length str) 0)
        (values '() '())
        (let ((func (if (equal? (string-ref str 0) #\<) sml-read-tag sml-read-text))
              (tag (if (equal? (string-ref str 0) #\<) 'tag 'text)))
          (let-values (((t r) (func str)))
            (values (cons tag t) r))))))

; Given a string of SML, returns a closure that returns one
; token each time it is evaluated. The closure returns '()
; if there are no more tokens.
(define make-sml-lexer
  (lambda (str)
    (let ((work str))
      (lambda ()
        (let-values (((p r) (sml-read-token work)))
          (if (equal? p '())
              '()
              (begin (set! work r) p)))))))

; Given a string of SML, returns all of its tokens as a list.
(define lex-sml
  (lambda (str)
    (let ((t (make-sml-lexer str)))
      (let reader ((a '()))
        (let ((n (t)))
          (if (null? n)
              (reverse a)
              (reader (cons n a))))))))

; Map of SML tags to parse tree symbols.
(define sml->symbol-alist
  '(("<B>" bold)
    ("<EM>" emph)
    ("<I>" ital)
    ("<PL>" plain)
    ("<S>" strong)
    ("<TT>" tt)
    ("<U>" u)
    ("<0>" font-0)
    ("<1>" font-1)
    ("<2>" font-2)
    ("<3>" font-3)
    ("<4>" font-4)
    ("<5>" font-5)
    ("<6>" font-6)
    ("<7>" font-7)
    ("<8>" font-8)
    ("<9>" font-9)
    ("<r>" color-red)
    ("<g>" color-green)
    ("<b>" color-blue)
    ("<c>" color-cyan)
    ("<m>" color-magenta)
    ("<y>" color-yellow)
    ("<k>" color-black)
    ("<w>" color-white)))

(define sml-font-code?
  (lambda (sym)
    (if (member sym '(font-0 font-1 font-2 font-3 font-4 font-5 font-6 font-7 font-8 font-9)) #t #f)))

(define sml-color-code?
  (lambda (sym)
    (if (member sym '(color-red color-green color-blue color-cyan color-magenta color-yellow color-black color-white)) #t #f)))

; Given a raw SML tag as a string (such as "<B>"), returns the
; corresponding parse tree symbol (in this case, 'bold).
; Returns #f if the symbol is unknown.
(define sml->symbol
  (lambda (str)
    (cadr (assoc str sml->symbol-alist))))

; Given a parse tree symbol, returns the corresponding SML tag
; as a string. Returns #f if the symbol is unknown.
(define symbol->sml
  (lambda (sym)
    (let scanner ((lst sml->symbol-alist))
      (if (null? lst)
          #f
          (if (equal? sym (cadar lst))
              (caar lst)
              (scanner (cdr lst)))))))

; Given an SML open tag as a string, generates the appropriate
; closing tag.
(define make-close-tag
  (lambda (str)
    (string-append "</" (substring str 1 (- (string-length str) 1)) ">")))

; Returns #t if the given tag (such as "</B>") is a close tag,
; or #f otherwise.
(define sml-close-tag?
  (lambda (str)
    (if (> (string-length str) 2)
        (if (equal? (string-ref str 1) #\/)
            #t
            #f)
        #f)))

; Converts an SML token to a parse tree node.
; For instance, turns '(tag "<B>") into '(bold).
(define recognize-sml-node
  (lambda (node)
    (if (equal? (car node) 'tag)
        (let ((item (assoc (cdr node) sml->symbol-alist)))
          (if (equal? item #f)
              (if (sml-close-tag? (cdr node))
                  'close
                  (error (format "Bad node: ~a" node)))
              (cadr item)))
        (cdr node))))

; Converts a well-formed SML document into a parse tree.
(define parse-sml
  (lambda (str)
    (if (string? str)
        (parse-sml (make-sml-lexer str))
        (let* ((lexer str)
               (token (lexer)))
          (if (null? token)
              '()
              (let ((node (recognize-sml-node token)))
                (cond ((equal? node 'close) '())
                      ((string? node)
                       (cons node (parse-sml lexer)))
                      ((symbol? node)
                       (let ((sub-tree (parse-sml lexer)))
                         (cons (cons node sub-tree) (parse-sml lexer)))))))))))

; Performs a simple whitespace removal pass; that is, removes
; redundant whitespace contained entirely within strings in which
; the TT attribute is not set.
(define remove-internal-whitespace
  (lambda (sml)
    (cond ((null? sml) '())
          ((list? (car sml))
           (case (caar sml)
             ((tt) (cons (car sml) (remove-internal-whitespace (cdr sml))))
             (else (cons (cons (caar sml) (remove-internal-whitespace (cdar sml)))
                         (remove-internal-whitespace (cdr sml))))))
          ((string? (car sml))
           (cons (remove-redundant-whitespace-string (car sml))
                 (remove-internal-whitespace (cdr sml))))
          (else (error "WTF?")))))

; Removes redundant whitespace from a string. Used by
; remove-internal-whitespace.
(define remove-redundant-whitespace-string
  (lambda (str)
    (let scanner ((in (string->list str))
                  (out '())
                  (ws #f))
      (cond ((null? in) (list->string (reverse out)))
            ((char-whitespace? (car in))
             (if ws
                 (scanner (cdr in) out #t)
                 (scanner (cdr in) (cons #\space out) #t)))
            (else (scanner (cdr in) (cons (car in) out) #f))))))

(define remove-redundant-tags-monkey
  (lambda (sml bold ital strong tt u size color)
    (cond ((null? sml) '())
          ((string? (car sml))
           (cons (car sml) (remove-redundant-tags-monkey (cdr sml) bold ital strong tt u size color)))
          ((list? (car sml))
           (let ((r (remove-redundant-tags-monkey (cdr sml) bold ital strong tt u size color)))
             (case (caar sml)
             ((bold) (if bold
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'bold (remove-redundant-tags-monkey  (cdar sml) #t ital strong tt u size color)) r)))
             ((ital) (if ital
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'ital (remove-redundant-tags-monkey (cdar sml) bold #t strong tt u size color)) r)))
             ((strong) (if strong
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'strong (remove-redundant-tags-monkey (cdar sml) bold ital #t tt u size color)) r)))
             ((tt) (if tt
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'tt (remove-redundant-tags-monkey (cdar sml) bold ital strong #t u size color)) r)))
             ((emph) (if strong
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'emph (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)))
             ((u) (if (>= u 3)
                         (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                         (cons (cons 'u (remove-redundant-tags-monkey (cdar sml) bold ital strong tt (+ u 1) size color)) r)))
             
             ((font-0 font-1 font-2 font-3 font-4 font-5 font-6 font-7 font-8 font-9)
              (if (equal? (caar sml) size)
                  (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                  (cons (cons (caar sml) (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u (caar sml) color)) r)))
             
             ((color-red color-green color-blue color-cyan color-magenta color-yellow color-black color-white)
              (if (equal? (caar sml) color)
                  (cons (cons 'none (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size color)) r)
                  (cons (cons (caar sml) (remove-redundant-tags-monkey (cdar sml) bold ital strong tt u size (caar sml))) r)))
             
             (else (cons (cons (caar sml) (remove-redundant-tags-monkey  (cdar sml) bold ital strong tt u size color)) r)))))
          (else (error "WTF?")))))

(define remove-redundant-tags
  (lambda (sml)
    (remove-redundant-tags-monkey sml #f #f #f #f 0 'unknown 'unknown)))

; Writes SML code representing the given parse tree to the
; current output port.
(define display-sml-from-tree
  (lambda (sml port)
    (if (not (null? sml))
        (cond ((string? (car sml))
               (display (car sml) port)
               (display-sml-from-tree (cdr sml) port))
              ((list? (car sml))
               (if (equal? (caar sml) 'none)
                   (display-sml-from-tree (cdar sml) port)
                   (let* ((open-tag (symbol->sml (caar sml)))
                          (close-tag (make-close-tag open-tag)))
                     (display open-tag port)
                     (display-sml-from-tree (cdar sml) port)
                     (display close-tag port)))
               (display-sml-from-tree (cdr sml) port))
              (else (error "WTF?"))))))

(define display-sml-from-list
  (lambda (sml port tag-stack)
    (if (not (null? sml))
        (cond ((string? (car sml))
               (display (car sml) port)
               (display-sml-from-list (cdr sml) port tag-stack))
              ((equal? (car sml) 'close)
               (display (make-close-tag (symbol->sml (car tag-stack))) port)
               (display-sml-from-list (cdr sml) port (cdr tag-stack)))
              ((symbol? (car sml))
               (display (symbol->sml (car sml)) port)
               (display-sml-from-list (cdr sml) port (cons (car sml) tag-stack)))
              (else (error "WTF?"))))))
               
; Generates a string of SML from a parse tree.
(define generate-sml-from-tree
  (lambda (sml)
    (let ((o (open-output-string)))
      (display-sml-from-tree sml o)
      (get-output-string o))))

; Generates a string of SML from a parsed list.
(define generate-sml-from-list
  (lambda (sml)
    (let ((o (open-output-string)))
      (display-sml-from-list sml o '())
      (get-output-string o))))

; Performs various optimizations on an SML string
; and returns the optimized SML code.
(define tree-optimize-sml-string
  (lambda (str)
    (let optimizer ((sml (parse-sml str))
                    (phases '(remove-internal-whitespace remove-redundant-tags)))
      (if (null? phases)
          (generate-sml-from-tree sml)
            (optimizer (eval (cons (car phases) (cons (list 'quote sml) '()))) (cdr phases))))))

(define listify-sml-tree
  (lambda (sml)
    (cond ((null? sml) '())
          ((string? (car sml)) (cons (car sml) (listify-sml-tree (cdr sml))))
          ((list? (car sml)) (append (list (caar sml)) (listify-sml-tree (cdar sml)) (list 'close) (listify-sml-tree (cdr sml))))
          (else (error "WTF?")))))

(define remove-first
  (lambda (lst item)
    (cond ((null? lst) '())
          ((equal? (car lst) item) (cdr lst))
          (else (cons (car lst) (remove-first (cdr lst) item))))))

(define remove-closing-tag
  (lambda (lst depth)
    (cond ((null? lst) '())
          ((string? (car lst))
           (cons (car lst) (remove-closing-tag (cdr lst) depth)))
          ((equal? (car lst) 'close)
           (if (= depth 1)
               (cdr lst)
               (cons (car lst) (remove-closing-tag (cdr lst) (- depth 1)))))
          ((symbol? (car lst))
           (cons (car lst) (remove-closing-tag (cdr lst) (+ depth 1))))
          (else (error (format "Bad list: ~a~n" lst))))))

(define find-within-branch
  (lambda (sml item depth)
    (cond ((null? sml) #f)
          ((equal? (car sml) item) #t)
          ((equal? (car sml) 'close)
           (if (= depth 0)
               #f
               (find-within-branch (cdr sml) item (- depth 1))))
          ((symbol? (car sml))
           (find-within-branch (cdr sml) item (+ depth 1)))
          (else (error "WTF?")))))

(define overlap-tags
  (lambda (sml pred)
    (cond ((null? sml) '())
          ((string? (car sml)) (cons (car sml) (overlap-tags (cdr sml) pred)))
          ((symbol? (car sml))
           ; if this is the right kind of tag...
           (if (pred (car sml))
               ; ...and it's used again elsewhere...
               (if (find-within-branch (cdr sml) (car sml) 1)
                   ; ... then try to overlap it.
                   ; to do this, remove its close tag and the next one's start tag.
                   (cons (car sml) (overlap-tags (remove-closing-tag (remove-first (cdr sml) (car sml)) 1) pred))
                   ; otherwise overlapping is of no use
                   (cons (car sml) (overlap-tags (cdr sml) pred)))
               (cons (car sml) (overlap-tags (cdr sml) pred))))
          (else (error (format "Bad list: ~a~n" sml))))))

(define overlap-colors
  (lambda (sml)
    (overlap-tags sml sml-color-code?)))

(define overlap-sizes
  (lambda (sml)
    (overlap-tags sml sml-font-code?)))

(define remove-close-open
  (lambda (sml)
    (cond ((null? sml) '())
          ((null? (cdr sml)) sml)
          ((equal? (car sml) 'close)
           (cons 'close (remove-close-open (cdr sml))))
          ((and (symbol? (car sml))
                (equal? (make-close-tag (symbol->sml (car sml))) (cadr sml)))
           (remove-close-open (cddr sml)))
          (else (cons (car sml) (remove-close-open (cdr sml)))))))

(define list-optimize-sml-string
  (lambda (str)
    (let optimizer ((sml (listify-sml-tree (parse-sml str)))
                    (phases '(remove-close-open)))
      (if (null? phases)
          (generate-sml-from-list sml)
          (optimizer (eval (cons (car phases) (cons (list 'quote sml) '()))) (cdr phases))))))

(define optimize-stdio
  (lambda ()
    (let ((str (let reader ((chars '()))
                 (let ((in (read-char)))
                   (if (eof-object? in)
                       (list->string (reverse chars))
                       (reader (cons in chars)))))))
      (let ((optimized (tree-optimize-sml-string (tree-optimize-sml-string str))))
        (if (< (string-length optimized) (string-length str))
            (display optimized)
            (display str))))))